import numpy as np
from sklearn.neighbors import BallTree, KDTree

from .base import IndexAdapter, register_default


@register_default("sklearn_kdtree")
class SklearnKDTreeAdapter(IndexAdapter):
    """Xoak index adapter for :class:`sklearn.neighbors.KDTree`."""

    def __init__(self, **kwargs):
        self._index_options = kwargs

    def build(self, points):
        return KDTree(points, **self._index_options)

    def query(self, kdtree, points):
        return kdtree.query(points)


@register_default("sklearn_balltree")
class SklearnBallTreeAdapter(IndexAdapter):
    """Xoak index adapter for :class:`sklearn.neighbors.BallTree`."""

    def __init__(self, **kwargs):
        self._index_options = kwargs

    def build(self, points):
        return BallTree(points, **self._index_options)

    def query(self, btree, points):
        return btree.query(points)


@register_default("sklearn_geo_balltree")
class SklearnGeoBallTreeAdapter(IndexAdapter):
    """Xoak index adapter for :class:`sklearn.neighbors.BallTree`, using
    the 'haversine' metric.

    It can be used for indexing a set of latitude / longitude points.

    When building the index, the coordinates must be given in the latitude,
    longitude order.

    Latitude and longitude values must be given in degrees for both index and
    query points (those values are converted in radians by this adapter).

    """

    def __init__(self, **kwargs):
        kwargs.update({"metric": "haversine"})
        self._index_options = kwargs

    def build(self, points):
        return BallTree(np.deg2rad(points), **self._index_options)

    def query(self, btree, points):
        return btree.query(np.deg2rad(points))



from importlib.metadata import version

from .accessor import XoakAccessor
from .index import IndexAdapter, IndexRegistry

__all__ = [
    "XoakAccessor",
    "IndexAdapter",
    "IndexRegistry",
]

__version__ = version("xoak")



import dask
import dask.array as da
import numpy as np
import pytest
import xarray as xr
from sklearn.metrics import pairwise_distances_argmin_min

# use single-threaded dask scheduler for all tests, as multi-threads or
# multi-processes may not be supported by some index adapters.
# TODO: enable multi-threaded and/or multi-processes per index
dask.config.set(scheduler="single-threaded")


@pytest.fixture(params=[np, da], scope="session")
def dataset_array_lib(request):
    """Array lib that is used for creation of the data."""
    return request.param


@pytest.fixture(params=[np, da], scope="session")
def indexer_array_lib(request):
    """Array lib that is used for creation of the indexer."""
    return request.param


@pytest.fixture(
    params=[
        (("d1",), (200,), (100,)),
        (("d1", "d2"), (20, 10), (10, 10)),
        (("d1", "d2", "d3"), (4, 10, 5), (2, 10, 5)),
    ],
    scope="session",
)
def dataset_dims_shape_chunks(request):
    return request.param


@pytest.fixture(
    params=[
        (("i1",), (100,), (50,)),
        (("i1", "i2"), (10, 10), (5, 10)),
        (("i1", "i2", "i3"), (2, 10, 5), (1, 10, 5)),
    ],
    scope="session",
)
def indexer_dims_shape_chunks(request):
    return request.param


def query_brute_force(
    dataset,
    dataset_dims_shape_chunks,
    indexer,
    indexer_dims_shape_chunks,
    metric="euclidean",
):
    """Find nearest neighbors using brute-force approach."""

    # for lat/lon coordinate, assume they are ordered lat, lon!!
    X = np.stack([np.ravel(c) for c in indexer.coords.values()]).T
    Y = np.stack([np.ravel(c) for c in dataset.coords.values()]).T

    if metric == "haversine":
        X = np.deg2rad(X)
        Y = np.deg2rad(Y)

    positions, _ = pairwise_distances_argmin_min(X, Y, metric=metric)

    dataset_dims, dataset_shape, _ = dataset_dims_shape_chunks
    indexer_dims, indexer_shape, _ = indexer_dims_shape_chunks

    u_positions = list(np.unravel_index(positions.ravel(), dataset_shape))

    pos_indexers = {
        dim: xr.Variable(indexer_dims, ind.reshape(indexer_shape))
        for dim, ind in zip(dataset_dims, u_positions)
    }

    return dataset.isel(indexers=pos_indexers)


@pytest.fixture(scope="session")
def geo_dataset(dataset_dims_shape_chunks, dataset_array_lib):
    """Dataset with coords lon and lat on a grid of different shapes."""
    dims, shape, chunks = dataset_dims_shape_chunks

    if dataset_array_lib is da:
        kwargs = {"size": shape, "chunks": chunks}
    else:
        kwargs = {"size": shape}

    lat = xr.DataArray(dataset_array_lib.random.uniform(-80, 80, **kwargs), dims=dims)
    lon = xr.DataArray(dataset_array_lib.random.uniform(-160, 160, **kwargs), dims=dims)

    ds = xr.Dataset(coords={"lat": lat, "lon": lon})

    return ds


@pytest.fixture(scope="session")
def geo_indexer(indexer_dims_shape_chunks, indexer_array_lib):
    """Indexer dataset with coords longitude and latitude of parametrized shapes."""
    dims, shape, chunks = indexer_dims_shape_chunks

    if indexer_array_lib is da:
        kwargs = {"size": shape, "chunks": chunks}
    else:
        kwargs = {"size": shape}

    latitude = xr.DataArray(indexer_array_lib.random.uniform(-80, 80, **kwargs), dims=dims)
    longitude = xr.DataArray(indexer_array_lib.random.uniform(-160, 160, **kwargs), dims=dims)

    ds = xr.Dataset(coords={"latitude": latitude, "longitude": longitude})

    return ds


@pytest.fixture(scope="session")
def geo_expected(geo_dataset, dataset_dims_shape_chunks, geo_indexer, indexer_dims_shape_chunks):
    return query_brute_force(
        geo_dataset,
        dataset_dims_shape_chunks,
        geo_indexer,
        indexer_dims_shape_chunks,
        metric="haversine",
    )


@pytest.fixture(scope="session")
def xyz_dataset(dataset_dims_shape_chunks, dataset_array_lib):
    """Dataset with coords x, y, z on a grid of different shapes."""
    dims, shape, chunks = dataset_dims_shape_chunks

    if dataset_array_lib is da:
        kwargs = {"size": shape, "chunks": chunks}
    else:
        kwargs = {"size": shape}

    x = xr.DataArray(dataset_array_lib.random.uniform(0, 10, **kwargs), dims=dims)
    y = xr.DataArray(dataset_array_lib.random.uniform(0, 10, **kwargs), dims=dims)
    z = xr.DataArray(dataset_array_lib.random.uniform(0, 10, **kwargs), dims=dims)

    ds = xr.Dataset(coords={"x": x, "y": y, "z": z})

    return ds


@pytest.fixture(scope="session")
def xyz_indexer(indexer_dims_shape_chunks, indexer_array_lib):
    """Indexer dataset with coords xx, yy, zz of parametrized shapes."""
    dims, shape, chunks = indexer_dims_shape_chunks

    if indexer_array_lib is da:
        kwargs = {"size": shape, "chunks": chunks}
    else:
        kwargs = {"size": shape}

    xx = xr.DataArray(indexer_array_lib.random.uniform(0, 10, **kwargs), dims=dims)
    yy = xr.DataArray(indexer_array_lib.random.uniform(0, 10, **kwargs), dims=dims)
    zz = xr.DataArray(indexer_array_lib.random.uniform(0, 10, **kwargs), dims=dims)

    ds = xr.Dataset(coords={"xx": xx, "yy": yy, "zz": zz})

    return ds


@pytest.fixture(scope="session")
def xyz_expected(xyz_dataset, dataset_dims_shape_chunks, xyz_indexer, indexer_dims_shape_chunks):
    return query_brute_force(
        xyz_dataset, dataset_dims_shape_chunks, xyz_indexer, indexer_dims_shape_chunks
    )



import pytest
import xarray as xr
from scipy.spatial import cKDTree

import xoak  # noqa: F401


def test_set_index_error():
    ds = xr.Dataset(
        coords={
            "x": (("a", "b"), [[0, 1], [2, 3]]),
            "y": (("b", "a"), [[0, 1], [2, 3]]),
            "z": ("a", [0, 1]),
        }
    )

    with pytest.raises(ValueError, match=".*must all have the same dimensions.*"):
        ds.xoak.set_index(["x", "y"], "scipy_kdtree")

    with pytest.raises(ValueError, match=".*must all have the same dimensions.*"):
        ds.xoak.set_index(["x", "z"], "scipy_kdtree")


def test_set_index_persist_false():
    ds = xr.Dataset(
        coords={
            "x": ("a", [0, 1, 2, 3]),
            "y": ("a", [0, 1, 2, 3]),
        }
    )
    ds = ds.chunk(2)

    ds.xoak.set_index(["x", "y"], "scipy_kdtree", persist=False)

    assert isinstance(ds.xoak._index, tuple)


def test_sel_error():
    ds = xr.Dataset(
        coords={
            "x": ("a", [0, 1, 2, 3]),
            "y": ("a", [0, 1, 2, 3]),
        }
    )
    indexer = xr.Dataset(
        coords={
            "x": ("p", [1.2, 2.9]),
            "y": ("p", [1.2, 2.9]),
        }
    )

    with pytest.raises(ValueError, match=".*not been built yet.*"):
        ds.xoak.sel(x=indexer.x, y=indexer.y)

    ds.xoak.set_index(["x", "y"], "scipy_kdtree")
    indexer = xr.Dataset(
        coords={
            "x": ("p", [1.2, 2.9]),
            "y": ("p2", [1.2, 2.9]),
        }
    )

    with pytest.raises(ValueError, match=".*must have the same dimensions.*"):
        ds.xoak.sel(x=indexer.x, y=indexer.y)


def test_index_property():
    ds = xr.Dataset(
        coords={
            "x": ("a", [0, 1, 2, 3]),
            "y": ("a", [0, 1, 2, 3]),
        }
    )

    assert ds.xoak.index is None

    ds.xoak.set_index(["x", "y"], "scipy_kdtree")
    assert isinstance(ds.xoak.index, cKDTree)

    ds_chunk = ds.chunk(2)
    ds_chunk.xoak.set_index(["x", "y"], "scipy_kdtree")
    assert isinstance(ds_chunk.xoak.index, list)



import sys

import numpy as np
import pytest
import xarray as xr

import xoak  # noqa:F401

pytest.importorskip("pys2index")


def test_s2point(geo_dataset, geo_indexer, geo_expected):
    geo_dataset.xoak.set_index(["lat", "lon"], "s2point")
    ds_sel = geo_dataset.xoak.sel(lat=geo_indexer.latitude, lon=geo_indexer.longitude)

    xr.testing.assert_equal(ds_sel.load(), geo_expected.load())


def test_s2point_sizeof():
    ds = xr.Dataset(coords={"lat": ("points", [0.0, 10.0]), "lon": ("points", [-5.0, 5.0])})
    points = np.array([[0.0, -5.0], [10.0, 5.0]])

    ds.xoak.set_index(["lat", "lon"], "s2point")

    assert sys.getsizeof(ds.xoak._index._index_adapter) > points.nbytes



from pys2index import S2PointIndex

from .base import IndexAdapter, register_default


@register_default("s2point")
class S2PointIndexAdapter(IndexAdapter):
    """Xoak index adapter for :class:`pys2index.S2PointIndex`.

    It can be used for efficient indexing of a set of latitude / longitude points.

    When building the index, the coordinates must be given in the latitude,
    longitude order.

    Latitude and longitude values must be in degrees for both index and query
    points.

    See https://github.com/benbovy/pys2index.

    """

    def build(self, points):
        self._points_nbytes = points.nbytes
        return S2PointIndex(points)

    def query(self, s2index, points):
        return s2index.query(points)

    def __sizeof__(self):
        # a very crude approx. of the index memory consumption, useful for Dask.
        # Unfortunately, we cannot get the actual size of the underlying index, as
        # the internal data structures used by the index are not exposed to Python.
        return self._points_nbytes



import numpy as np
import pytest

from xoak import IndexAdapter, IndexRegistry
from xoak.index.base import (
    IndexRegistrationWarning,
    XoakIndexWrapper,
    normalize_index,
    register_default,
)
from xoak.index.scipy_adapters import ScipyKDTreeAdapter


class DummyIndex:
    def __init__(self, points, option=1):
        self.points = points
        self.option = option

    def query(self, points):
        distances = np.zeros(points.shape[0])
        indices = np.ones(points.shape[0])

        return distances, indices


class DummyIndexAdapter(IndexAdapter):
    def __init__(self, **kwargs):
        self.index_kwargs = kwargs

    def build(self, points):
        return DummyIndex(points, **self.index_kwargs)

    def query(self, index, points):
        return index.query(points)


def test_index_adapter_base():
    class IndexAdapterSubclass(IndexAdapter):
        def build(self, points):
            return super().build(points)

        def query(self, index, points):
            return super().query(index, points)

    adapter = IndexAdapterSubclass()

    with pytest.raises(NotImplementedError):
        adapter.build(np.zeros((10, 2)))

    with pytest.raises(NotImplementedError):
        adapter.query(None, np.zeros((10, 2)))


def test_index_registery_constructor():
    registry = IndexRegistry()
    assert dict(registry) == registry._default_indexes

    registry = IndexRegistry(use_default=False)
    assert len(registry) == 0


def test_index_registery_register():
    registry = IndexRegistry(use_default=False)
    registry.register("dummy")(DummyIndexAdapter)

    with pytest.warns(IndexRegistrationWarning, match="overriding an already registered index.*"):
        registry.register("dummy")(DummyIndexAdapter)

    with pytest.raises(TypeError, match="can only register IndexAdapter subclasses."):
        registry.register("invalid")(DummyIndex)


def test_index_registry_dict_interface():
    registry = IndexRegistry(use_default=False)
    registry.register("dummy")(DummyIndexAdapter)

    assert registry["dummy"] is DummyIndexAdapter
    assert list(registry) == ["dummy"]
    assert len(registry) == 1
    assert repr(registry) == "<IndexRegistry (1 indexes)>\ndummy"


def test_index_registry_attr_access():
    registry = IndexRegistry(use_default=False)
    registry.register("dummy")(DummyIndexAdapter)

    assert registry.dummy is DummyIndexAdapter
    assert "dummy" in dir(registry)

    with pytest.raises(AttributeError, match=".*has no attribute.*"):
        registry.invalid_attr

    with pytest.raises(AttributeError, match=".*cannot set attribute.*"):
        registry.custom = DummyIndexAdapter


def test_index_registry_ipython_completion():
    registry = IndexRegistry(use_default=False)
    registry.register("dummy")(DummyIndexAdapter)

    assert "dummy" in registry._ipython_key_completions_()


def test_register_default():
    # check that docstrings are updated
    assert "This index adapter is registered in xoak" in ScipyKDTreeAdapter.__doc__

    register_default("dummy")(DummyIndexAdapter)
    assert "This index adapter is registered in xoak" in DummyIndexAdapter.__doc__
    del IndexRegistry._default_indexes["dummy"]


def test_normalize_index():
    assert normalize_index(DummyIndexAdapter) is DummyIndexAdapter
    assert normalize_index("scipy_kdtree") is ScipyKDTreeAdapter

    with pytest.raises(TypeError, match=".*is not a subclass of IndexAdapter"):
        normalize_index(DummyIndex)


def test_xoak_index_wrapper():
    idx_points = np.zeros((10, 2))
    offset = 1

    wrapper = XoakIndexWrapper(DummyIndexAdapter, idx_points, offset, option=2)
    wrapper2 = XoakIndexWrapper(DummyIndexAdapter, idx_points, 0)

    assert isinstance(wrapper.index, DummyIndex)
    assert wrapper.index.option == 2
    assert isinstance(wrapper2.index, DummyIndex)
    assert wrapper2.index.option == 1

    results = wrapper.query(np.zeros((5, 2))).ravel()

    assert results["distances"].dtype == np.double
    assert results["indices"].dtype == np.intp
    np.testing.assert_equal(results["distances"], np.zeros(5))
    np.testing.assert_equal(results["indices"], np.ones(5) + offset)



import pytest
import xarray as xr

import xoak  # noqa: F401

pytest.importorskip("scipy")


def test_scipy_kdtree(xyz_dataset, xyz_indexer, xyz_expected):
    xyz_dataset.xoak.set_index(["x", "y", "z"], "scipy_kdtree")
    ds_sel = xyz_dataset.xoak.sel(x=xyz_indexer.xx, y=xyz_indexer.yy, z=xyz_indexer.zz)

    xr.testing.assert_equal(ds_sel.load(), xyz_expected.load())


def test_scipy_kdtree_options():
    ds = xr.Dataset(coords={"x": ("points", [1, 2]), "y": ("points", [1, 2])})

    ds.xoak.set_index(["x", "y"], "scipy_kdtree", leafsize=10)

    assert ds.xoak.index.leafsize == 10



import importlib

from .base import IndexAdapter, IndexRegistry  # noqa: F401

adapters = [
    "scipy_adapters",
    "sklearn_adapters",
    "s2_adapters",
]

for mod in adapters:
    try:
        # importing the module registers the adapters
        importlib.import_module("." + mod, package="xoak.index")
    except ImportError:  # pragma: no cover
        pass

del adapters



from scipy.spatial import cKDTree

from .base import IndexAdapter, register_default


@register_default("scipy_kdtree")
class ScipyKDTreeAdapter(IndexAdapter):
    """Xoak index adapter for :class:`scipy.spatial.cKDTree`."""

    def __init__(self, **kwargs):
        self.index_options = kwargs

    def build(self, points):
        return cKDTree(points, **self.index_options)

    def query(self, kdtree, points):
        return kdtree.query(points)



from __future__ import annotations

import abc
import warnings
from collections.abc import Mapping
from contextlib import suppress
from typing import Any, TypeVar

import numpy as np

Index = TypeVar("Index")


class IndexAdapter(abc.ABC):
    """Base class for reusing a custom index to select data in
    :class:`xarray.DataArray` or :class:`xarray.Dataset` objects with xoak.

    Subclasses must implement the ``build()`` and ``query()`` methods,
    which are called to build a new index and query this index, respectively.

    If any options are necessary, they should be implemented as arguments to the
    ``__init__()`` method.

    """

    def __init__(self, **kwargs):
        pass

    @abc.abstractmethod
    def build(self, points: np.ndarray) -> Index:
        """Build the index from a set of points/samples and their coordinate labels.

        Parameters
        ----------
        points : ndarray of shape (n_points, n_coordinates)
            Two-dimensional array of points/samples (rows) and their
            corresponding coordinate labels (columns) to index.

        Returns
        -------
        index: object
            A new index object.

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def query(self, index: Index, points: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """Query points/samples,

        Parameters
        ----------
        index: object
            The index object returned by ``build()``.
        points: ndarray of shape (n_points, n_coordinates)
            Two-dimensional array of points/samples (rows) and their
            corresponding coordinate labels (columns) to query.

        Returns
        -------
        distances : ndarray of shape (n_points)
            Distances to the nearest neighbors.
        indices : ndarray of shape (n_points)
            Indices of the nearest neighbors in the array of the indexed
            points.

        """
        raise NotImplementedError()


class IndexRegistrationWarning(Warning):
    """Warning for conflicts in index registration."""


class IndexRegistry(Mapping[str, type[IndexAdapter]]):
    """A registry of all indexes adapters that can be used to select data
    with xoak.

    """

    _default_indexes: dict[str, type[IndexAdapter]] = {}

    def __init__(self, use_default=True):
        """Creates a new index registry.

        This registry provides a dict-like interface as well as attribute-style
        access to index adapters.

        Parameters
        ----------
        use_default : bool, optional
            If True (default), pre-populates the registry with xoak's built-in
            index adapters.

        """
        self._indexes = {}

        if use_default:
            self._indexes.update(self._default_indexes)

    def register(self, name: str):
        """Register custom index in xoak.

        Parameters
        ----------
        name : str
            Name to give to this index type.
        cls: :class:`IndexAdapter` subclass
            The index adapter class to register.

        """

        def wrap(cls: type[IndexAdapter]):
            if not issubclass(cls, IndexAdapter):
                raise TypeError("can only register IndexAdapter subclasses.")

            if name in self._indexes:
                warnings.warn(
                    f"overriding an already registered index with the name '{name}'.",
                    IndexRegistrationWarning,
                    stacklevel=2,
                )

            self._indexes[name] = cls

            return cls

        return wrap

    def __getattr__(self, name):
        if name not in {"__dict__", "__setstate__"}:
            # this avoids an infinite loop when pickle looks for the
            # __setstate__ attribute before the xarray object is initialized
            with suppress(KeyError):
                return self._indexes[name]
        raise AttributeError(f"IndexRegistry object has no attribute {name!r}")

    def __setattr__(self, name, value):
        if name == "_indexes":
            object.__setattr__(self, name, value)
        else:
            raise AttributeError(
                f"cannot set attribute {name!r} on a IndexRegistry object. "
                "Use `.register()` to add a new index adapter to the registry."
            )

    def __dir__(self):
        extra_attrs = [k for k in self._indexes]
        return sorted(set(dir(type(self)) + extra_attrs))

    def _ipython_key_completions_(self):
        return list(self._indexes)

    def __getitem__(self, key):
        return self._indexes[key]

    def __iter__(self):
        return iter(self._indexes)

    def __len__(self):
        return len(self._indexes)

    def __repr__(self):
        header = f"<IndexRegistry ({len(self._indexes)} indexes)>\n"
        return header + "\n".join([name for name in self._indexes])


def register_default(name: str):
    """A convenient decorator to register xoak's builtin indexes."""

    doc_extra = f"""

    This index adapter is registered in xoak under the name ``{name}``.
    You can use it in :meth:`xarray.Dataset.xoak.set_index` by simply providing
    its name for the ``index_type`` argument.
    Alternatively, you can access it via the index registry, i.e.,

    >>> import xoak
    >>> ireg = xoak.IndexRegistry()
    >>> ireg.{name}

    """

    def decorator(cls: type[IndexAdapter]):
        if cls.__doc__ is not None:
            cls.__doc__ += doc_extra
        else:
            cls.__doc__ = doc_extra

        IndexRegistry._default_indexes[name] = cls
        return cls

    return decorator


def normalize_index(name_or_cls: str | Any) -> type[IndexAdapter]:
    if isinstance(name_or_cls, str):
        cls = IndexRegistry._default_indexes[name_or_cls]
    else:
        cls = name_or_cls

    if not issubclass(cls, IndexAdapter):
        raise TypeError(f"'{name_or_cls}' is not a subclass of IndexAdapter")

    return cls


class XoakIndexWrapper:
    """Thin wrapper used internally to build and query (registered)
    indexes, with dask support.

    """

    _query_result_dtype: list[tuple[str, Any]] = [
        ("distances", np.double),
        ("indices", np.intp),
    ]

    def __init__(
        self,
        index_adapter: str | type[IndexAdapter],
        points: np.ndarray,
        offset: int,
        **kwargs,
    ):
        index_adapter_cls = normalize_index(index_adapter)

        self._index_adapter = index_adapter_cls(**kwargs)
        self._index = self._index_adapter.build(points)
        self._offset = offset

    @property
    def index(self):
        return self._index

    def query(self, points: np.ndarray) -> np.ndarray:
        distances, positions = self._index_adapter.query(self._index, points)

        result = np.empty(shape=points.shape[0], dtype=self._query_result_dtype)
        result["distances"] = distances.ravel().astype(np.double)
        result["indices"] = positions.ravel().astype(np.intp) + self._offset

        return result[:, None]



from __future__ import annotations

from collections.abc import Hashable, Iterable, Mapping
from typing import Any

import numpy as np
import xarray as xr
from xarray.core.utils import either_dict_or_kwargs

from .index.base import Index, IndexAdapter, XoakIndexWrapper

try:
    from dask.delayed import Delayed
except ImportError:  # pragma: no cover
    Delayed = type[None]


def coords_to_point_array(coords: list[Any]) -> np.ndarray:
    """Re-arrange data from a list of xarray coordinates into a 2-d array of shape
    (npoints, ncoords).

    """
    c_chunks = [c.chunks for c in coords]

    if any([chunks is None for chunks in c_chunks]):
        # plain numpy arrays (maybe triggers compute)
        X = np.stack([np.ravel(c) for c in coords]).T

    else:
        import dask.array as da

        # TODO: check chunks are equal for all coords?

        X = da.stack([da.ravel(c.data) for c in coords]).T
        X = X.rechunk((X.chunks[0], len(coords)))

    return X


IndexAttr = XoakIndexWrapper | Iterable[XoakIndexWrapper] | Iterable[Delayed]
IndexType = str | type[IndexAdapter]


@xr.register_dataarray_accessor("xoak")
@xr.register_dataset_accessor("xoak")
class XoakAccessor:
    """A xarray Dataset or DataArray extension for indexing irregular,
    n-dimensional data using a ball tree.

    """

    _index: IndexAttr
    _index_type: IndexType
    _index_coords: tuple[str, ...]
    _index_coords_dims: tuple[Hashable, ...]
    _index_coords_shape: tuple[int, ...]

    def __init__(self, xarray_obj: xr.Dataset | xr.DataArray):
        self._xarray_obj = xarray_obj

    def _build_index_forest_delayed(self, X, persist=False, **kwargs) -> IndexAttr:
        import dask

        indexes = []
        offset = 0

        for i, chunk in enumerate(X.to_delayed().ravel()):
            indexes.append(
                dask.delayed(XoakIndexWrapper)(self._index_type, chunk, offset, **kwargs)
            )
            offset += X.chunks[0][i]

        if persist:
            return dask.persist(*indexes)
        else:
            return tuple(indexes)

    def set_index(
        self,
        coords: Iterable[str],
        index_type: IndexType,
        persist: bool = True,
        **kwargs,
    ):
        """Create an index tree from a subset of coordinates of the DataArray / Dataset.

        If the given coordinates are chunked (Dask arrays), this method will (lazily) create
        a forest of index trees (one tree per chunk of the flattened coordinate arrays).

        Parameters
        ----------
        coords : iterable
            Coordinate names. Each given coordinate must have the same dimension(s),
            in the same order.
        index_type : str or :class:`~xoak.IndexAdapter` subclass
            Either a registered index adapter (see :class:`~xoak.IndexRegistry`) or a custom
            :class:`~xoak.IndexAdapter` subclass.
        persist: bool
            If True (default), this method will precompute and persist in memory the forest
            of index trees, if any.
        **kwargs
            Keyword arguments that will be passed to the underlying index constructor.

        """
        self._index_type = index_type
        self._index_coords = tuple(coords)

        coord_objs = [self._xarray_obj.coords[cn] for cn in coords]

        if len(set([c.dims for c in coord_objs])) > 1:
            raise ValueError(
                "Coordinates {coords} must all have the same dimensions in the same order"
            )

        self._index_coords_dims = coord_objs[0].dims
        self._index_coords_shape = coord_objs[0].shape

        X = coords_to_point_array([self._xarray_obj[c] for c in coords])

        if isinstance(X, np.ndarray):
            self._index = XoakIndexWrapper(self._index_type, X, 0, **kwargs)
        else:
            self._index = self._build_index_forest_delayed(X, persist=persist, **kwargs)

    @property
    def index(self) -> None | Index | Iterable[Index]:
        """Returns the underlying index object(s), or ``None`` if no index has
        been set yet.

        May trigger computation of lazy indexes.

        """
        if not getattr(self, "_index", False):
            return None
        elif isinstance(self._index, XoakIndexWrapper):
            return self._index.index
        else:
            import dask

            index_wrappers = dask.compute(*self._index)
            return [wrp.index for wrp in index_wrappers]

    def _query(self, indexers):
        X = coords_to_point_array([indexers[c] for c in self._index_coords])

        if isinstance(X, np.ndarray) and isinstance(self._index, XoakIndexWrapper):
            # directly call index wrapper's query method
            res = self._index.query(X)
            results = res["indices"][:, 0]

        else:
            # Two-stage lazy query with dask
            import dask
            import dask.array as da

            # coerce query array as a dask array and index(es) as an iterable
            if isinstance(X, np.ndarray):
                X = da.from_array(X, chunks=X.shape)

            if isinstance(self._index, XoakIndexWrapper):
                indexes = [self._index]
            else:
                indexes = self._index

            # 1st "map" stage:
            # - execute `IndexWrapperCls.query` for each query array chunk and each index instance
            # - concatenate all distances/positions results in two dask arrays of shape
            #   (n_points, n_indexes)

            res_chunk = []

            for i, chunk in enumerate(X.to_delayed().ravel()):
                res_chunk_idx = []

                chunk_npoints = X.chunks[0][i]
                shape = (chunk_npoints, 1)

                for idx in indexes:
                    dlyd = dask.delayed(idx.query)(chunk)
                    res_chunk_idx.append(
                        da.from_delayed(dlyd, shape, dtype=XoakIndexWrapper._query_result_dtype)
                    )

                res_chunk.append(da.concatenate(res_chunk_idx, axis=1))

            map_results = da.concatenate(res_chunk, axis=0)
            distances = map_results["distances"]
            indices = map_results["indices"]

            # 2nd "reduce" stage:
            # - brute force lookup over the indexes dimension (columns)

            indices_col = da.argmin(distances, axis=1)

            results = da.blockwise(
                lambda arr, icol: np.take_along_axis(arr, icol[:, None], 1).squeeze(),
                "i",
                indices,
                "ik",
                indices_col,
                "i",
                dtype=np.intp,
                concatenate=True,
            )

        return results

    def _get_pos_indexers(self, indices, indexers):
        """Returns positional indexers based on the query results and the
        original (label-based) indexers.

        1. Unravel the (flattened) indices returned from the query
        2. Reshape the unraveled indices according to indexers shapes
        3. Wrap the indices in xarray.Variable objects.

        """
        pos_indexers = {}

        indexer_dims = [idx.dims for idx in indexers.values()]
        indexer_shapes = [idx.shape for idx in indexers.values()]

        if len(set(indexer_dims)) > 1:
            raise ValueError("All indexers must have the same dimensions.")

        u_indices = list(np.unravel_index(indices.ravel(), self._index_coords_shape))

        for dim, ind in zip(self._index_coords_dims, u_indices):
            pos_indexers[dim] = xr.Variable(
                indexer_dims[0],
                ind.reshape(indexer_shapes[0]),
            )

        return pos_indexers

    def sel(
        self, indexers: Mapping[Hashable, Any] | None = None, **indexers_kwargs: Any
    ) -> xr.Dataset | xr.DataArray:
        """Selection based on a ball tree index.

        The index must have been already built using `xoak.set_index()`.

        It behaves mostly like :meth:`xarray.Dataset.sel` and
        :meth:`xarray.DataArray.sel` methods, with some limitations:

        - Orthogonal indexing is not supported
        - For vectorized (point-wise) indexing, you need to supply xarray
          objects
        - Use it for nearest neighbor lookup only (it implicitly
          assumes method="nearest")

        This triggers :func:`dask.compute` if the given indexers and/or the index
        coordinates are chunked.

        """
        if not getattr(self, "_index", False):
            raise ValueError(
                "The index(es) has/have not been built yet. Call `.xoak.set_index()` first"
            )

        indexers = either_dict_or_kwargs(indexers, indexers_kwargs, "xoak.sel")
        indices = self._query(indexers)

        if not isinstance(indices, np.ndarray):
            # TODO: remove (see todo below)
            indices = indices.compute()

        pos_indexers = self._get_pos_indexers(indices, indexers)

        # TODO: issue in xarray. 1-dimensional xarray.Variables are always considered
        # as OuterIndexer, while we want here VectorizedIndexer
        # This would also allow lazy selection
        result = self._xarray_obj.isel(indexers=pos_indexers)

        return result



import pytest
import xarray as xr

import xoak  # noqa: F401

pytest.importorskip("sklearn")


def test_sklearn_kdtree(xyz_dataset, xyz_indexer, xyz_expected):
    xyz_dataset.xoak.set_index(["x", "y", "z"], "sklearn_kdtree")
    ds_sel = xyz_dataset.xoak.sel(x=xyz_indexer.xx, y=xyz_indexer.yy, z=xyz_indexer.zz)

    xr.testing.assert_equal(ds_sel.load(), xyz_expected.load())


def test_sklearn_kdtree_options():
    ds = xr.Dataset(coords={"x": ("points", [1, 2]), "y": ("points", [1, 2])})

    ds.xoak.set_index(["x", "y"], "sklearn_kdtree", leaf_size=10)

    # sklearn tree classes init options are not exposed as class properties
    assert ds.xoak._index._index_adapter._index_options == {"leaf_size": 10}


def test_sklearn_balltree(xyz_dataset, xyz_indexer, xyz_expected):
    xyz_dataset.xoak.set_index(["x", "y", "z"], "sklearn_balltree")
    ds_sel = xyz_dataset.xoak.sel(x=xyz_indexer.xx, y=xyz_indexer.yy, z=xyz_indexer.zz)

    xr.testing.assert_equal(ds_sel.load(), xyz_expected.load())


def test_sklearn_balltree_options():
    ds = xr.Dataset(coords={"x": ("points", [1, 2]), "y": ("points", [1, 2])})

    ds.xoak.set_index(["x", "y"], "sklearn_balltree", leaf_size=10)

    # sklearn tree classes init options are not exposed as class properties
    assert ds.xoak._index._index_adapter._index_options == {"leaf_size": 10}


def test_sklearn_geo_balltree(geo_dataset, geo_indexer, geo_expected):
    geo_dataset.xoak.set_index(["lat", "lon"], "sklearn_geo_balltree")
    ds_sel = geo_dataset.xoak.sel(lat=geo_indexer.latitude, lon=geo_indexer.longitude)

    xr.testing.assert_equal(ds_sel.load(), geo_expected.load())


def test_sklearn_geo_balltree_options():
    ds = xr.Dataset(coords={"x": ("points", [1, 2]), "y": ("points", [1, 2])})

    ds.xoak.set_index(["x", "y"], "sklearn_geo_balltree", leaf_size=10, metric="euclidean")

    # sklearn tree classes init options are not exposed as class properties
    # user-defined metric should be ignored
    assert ds.xoak._index._index_adapter._index_options == {
        "leaf_size": 10,
        "metric": "haversine",
    }
